#!/usr/bin/env python

#
# Copyright (c) 2019 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND VMWARE DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VMWARE BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#

"""Following tests are covered to test route-map functionality.
TC_57:
    Create route map to match prefix-list and permit inbound
    and outbound prefixes and set criteria on match
TC_52:
    Test modify set/match clauses in a route-map to see
    if it takes immediate effect.
TC_61:
    Delete the route maps.
TC_50_1:
    Test modify/remove prefix-lists referenced by a
    route-map for match statement.
TC_50_1:
    Remove prefix-list referencec by route-map match cluase
    and verifying it reflecting as intended
TC_51:
    Add and remove community-list referencec by route-map match cluase
    and verifying it reflecting as intended
TC_45:
    Test multiple match statements as part of a route-map's single
    sequence number. (Logical OR-ed of multiple match statements)
TC_44:
    Test multiple match statements as part of a route-map's single
    sequence number. (Logical AND of multiple match statements)
TC_41:
    Test add/remove route-maps to specific neighbor and see if
    it takes effect as intended
TC_56:
    Test clear BGP sessions and interface flaps to see if
    route-map properties are intact.
TC_46:
    Verify if a blank sequence number can be create(without any
    match/set clause) and check if it allows all the traffic/prefixes
TC_48:
    Create route map setting local preference and weight to eBGP peeer
    and metric to ibgp peer and verifying it should not get advertised
TC_43:
    Test multiple set statements as part of a route-map's
    single sequence number.
TC_54:
    Verify route-maps continue clause functionality.
TC_55:
    Verify route-maps goto clause functionality.
TC_53:
    Verify route-maps call clause functionality.
TC_58:
    Create route map deny inbound and outbound prefixes on
    match prefix list and set criteria on match
TC_59:
    Create route map to permit inbound prefixes with filter
    match tag and set criteria
TC_60
    Create route map to deny outbound prefixes with filter match tag,
    and set criteria
"""

#################################
# TOPOLOGY
#################################
"""

                    +-------+
         +--------- |  R2   |
         |          +-------+
         |iBGP           |
     +-------+           |
     |  R1   |           |iBGP
     +-------+           |
         |               |
         |    iBGP   +-------+   eBGP   +-------+
         +---------- |  R3   |----------|  R4   |
                     +-------+          +-------+
                        |
                        |eBGP
                        |
                    +-------+
                    |  R5   |
                    +-------+


"""

import sys
import json
import time
import pytest
import inspect
import os
from time import sleep

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, '../'))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, get_topogen
from mininet.topo import Topo

# Required to instantiate the topology builder class.
from lib.common_config import (
    start_topology,  write_test_header,
    write_test_footer, create_static_routes,
    verify_rib, delete_route_maps, create_bgp_community_lists,
    interface_status, create_route_maps, create_prefix_lists,
    verify_route_maps, check_address_types, verify_bgp_community,
    shutdown_bringup_interface, verify_prefix_lists, reset_config_on_routers)
from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence, create_router_bgp,
    clear_bgp_and_verify, verify_bgp_attributes)
from lib.topojson import build_topo_from_json, build_config_from_json

# Reading the data from JSON File for topology and configuration creation
jsonFile = "{}/bgp_route_map_topo2.json".format(CWD)

try:
    with open(jsonFile, 'r') as topoJson:
        topo = json.load(topoJson)
except IOError:
    assert False, "Could not read file {}".format(jsonFile)

# Global variables
# Global variables
bgp_convergence = False
NETWORK = {
    "ipv4": ["11.0.20.1/32", "11.0.20.2/32"],
    "ipv6": ["2::1/128", "2::2/128"]
}

bgp_convergence = False
BGP_CONVERGENCE = False
ADDR_TYPES = check_address_types()


class BGPRmapTopo(Topo):
    """BGPRmapTopo.

    BGPRmap topology 1
    * `Topo`: Topology object
    """

    def build(self, *_args, **_opts):
        """Build function."""
        tgen = get_topogen(self)

        # Building topology and configuration from json file
        build_topo_from_json(tgen, topo)


def setup_module(mod):
    """setup_module.

    Set up the pytest environment
    * `mod`: module name
    """
    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("="*40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    tgen = Topogen(BGPRmapTopo, mod.__name__)
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start deamons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Checking BGP convergence
    global bgp_convergence
    global ADDR_TYPES

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Api call verify whether BGP is converged
    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, ('setup_module :Failed \n Error:'
                                     ' {}'.format(bgp_convergence))
    logger.info("Running setup_module() done")


def teardown_module(mod):
    """teardown_module.

    Teardown the pytest environment.
    * `mod`: module name
    """
    logger.info("Running teardown_module to delete topology")
    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info("Testsuite end time: {}".format(
                    time.asctime(time.localtime(time.time()))))
    logger.info("="*40)


#####################################################
# Tests starting
#####################################################


def test_rmap_match_prefix_list_permit_in_and_outbound_prefixes_p0():
    """
    TC: 57
    Create route map to match prefix-list and permit inbound
    and outbound prefixes and set criteria on match
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                    'pf_list_1_ipv4': [{
                        'seqid': 10,
                        'network': 'any',
                        'action': 'permit',
                    }]
                },
                'ipv6': {
                    'pf_list_1_ipv6': [{
                        'seqid': 10,
                        'network': 'any',
                        'action': 'permit',
                    }]
                }
            }
        }
    }

    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)
    for addr_type in ADDR_TYPES:
    # Create route map
        input_dict_3 = {
            "r3": {
                "route_maps": {
                    "rmap_match_pf_1_{}".format(addr_type): [{
                        "action": "permit",
                        'seq_id': '5',
                        "match": {
                            addr_type: {
                                "prefix_lists": "pf_list_1_" + addr_type
                            }
                        },
                        "set": {
                            "localpref": 150,
                            "weight": 100
                        }
                    },
                    ],
                    "rmap_match_pf_2_{}".format(addr_type): [{
                        "action": "permit",
                        'seq_id': '5',
                        "match": {
                            addr_type: {
                                "prefix_lists": "pf_list_1_" + addr_type
                            }
                        },
                        "set": {
                            "med": 50
                        }
                    },
                    ]
                }
            }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
           "bgp": {
               "address_family": {
                   "ipv4": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv4",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               },
                               "r4": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_2_ipv4",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   },
                   "ipv6": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv6",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               },
                               "r4": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_2_ipv6",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
       }
    }

    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]

    # dual stack changes
    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result4 is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result4)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    # dual stack changes
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result4 = verify_bgp_attributes(tgen, addr_type, dut, routes[
            addr_type],rmap_name, input_dict_3)
        assert result4 is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result4)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    # dual stack changes
    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result4 is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result4)

    # Verifying BGP set attributes
    dut = 'r4'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    # dual stack changes
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_2_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)
    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_modify_set_match_clauses_in_rmap_p0():
    """
    TC_52:
    Test modify set/match clauses in a route-map to see
    if it takes immediate effect.
    """

    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list

    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                    'pf_list_1_ipv4': [{
                        'seqid': 10,
                        'network': 'any',
                        'action': 'permit',
                    }],
                    'pf_list_2_ipv4': [{
                        'seqid': 10,
                        'network': 'any',
                        'action': 'permit'
                    }]
                },
                'ipv6': {
                    'pf_list_1_ipv6': [{
                        'seqid': 10,
                        'network': 'any',
                        'action': 'permit',
                    }],
                    'pf_list_2_ipv6': [{
                        'seqid': 10,
                        'network': 'any',
                        'action': 'permit'
                    }]
                }
            }
        }
        }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                    },
                    "set": {
                        "localpref": 150,
                    }
                }],
                "rmap_match_pf_2_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                    },
                    "set": {
                        "med": 50
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
           "bgp": {
               "address_family": {
                   "ipv4": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv4",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               },
                               "r4": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_2_ipv4",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   },
                   "ipv6": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv6",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               },
                               "r4": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_2_ipv6",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
       }
   }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    # dual stack changes
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result4 = verify_bgp_attributes(tgen, addr_type, dut, routes[
            addr_type],rmap_name, input_dict_3)
        assert result4 is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result4)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    # dual stack changes
    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result4 is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result4)

    # Verifying BGP set attributes
    dut = 'r4'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_2_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[
            addr_type],rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Modify set/match clause of in-used route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
        "route_maps": {
            "rmap_match_pf_1_{}".format(addr_type): [{
                "action": "permit",
                'seq_id': '5',
                "match": {
                    addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                },
                "set": {
                    "localpref": 1000,
                }
            }],
            "rmap_match_pf_2_{}".format(addr_type): [{
                "action": "permit",
                'seq_id': '5',
                "match": {
                    addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                },
                "set": {
                    "med": 2000
                }
            }]
        }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying BGP set attributes
    dut = 'r4'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_2_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_delete_route_maps_p1():
    """
    TC_61:
    Delete the route maps.
    """

    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_tag_1_{}".format(addr_type): [{
                    "action": "deny",
                    "match": {
                        addr_type: {
                        "tag": "4001"
                    }
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Delete route maps
    for addr_type in ADDR_TYPES:
        input_dict = {
        'r3': {
            'route_maps': ['rmap_match_tag_1_{}'.format(addr_type)]
        }
        }
        result = delete_route_maps(tgen, input_dict)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    result = verify_route_maps(tgen, input_dict)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)
    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_modify_prefix_list_referenced_by_rmap_p0():
    """
    TC_50_1:
    Test modify/remove prefix-lists referenced by a
    route-map for match statement.
    """

    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                    'pf_list_1_ipv4': [{
                        'seqid': 10,
                        'network': 'any',
                        'action': 'permit',
                    }]
                },
                'ipv6': {
                    'pf_list_1_ipv6': [{
                        'seqid': 100,
                        'network': 'any',
                        'action': 'permit',
                    }]
                }
            }
        }
        }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                    },
                    "set": {
                        "localpref": 150,
                        "weight": 100
                    }
                }],
                "rmap_match_pf_2_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                    },
                    "set": {
                        "med": 50
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
           "bgp": {
               "address_family": {
                   "ipv4": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv4",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               },
                               "r4": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_2_ipv4",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   },
                   "ipv6": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv6",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               },
                               "r4": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_2_ipv6",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
       }
   }

    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r4'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }

    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_2_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Modify ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'deny'
                }]
            },
            'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'deny'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    sleep(5)
    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
                protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: routes are not present \n '
        'Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
                protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: routes are not present \n '
        'Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_remove_prefix_list_referenced_by_rmap_p0():
    """
    TC_50_1:
    Remove prefix-list referencec by route-map match cluase
    and verifying it reflecting as intended
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
        }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                    },
                    "set": {
                        "localpref": 150,
                    }
                }],
                "rmap_match_pf_2_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "med": 50
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
       'r3': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv4",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 },
                                 "r4": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_2_ipv4",
                                                     "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                    "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv6",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 },
                                 "r4": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_2_ipv6",
                                                     "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
     }
        result = create_router_bgp(tgen, topo, input_dict_4)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r4'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_2_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                        rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Remove/Delete prefix list
    input_dict_3 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit',
                    'delete': True
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit',
                    'delete': True
                }]
            }
        }
        }
    }
    result = create_prefix_lists(tgen, input_dict_3)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    result = verify_prefix_lists(tgen, input_dict_3)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Api call to clear bgp, so config changes would be reflected
    dut = 'r3'
    result = clear_bgp_and_verify(tgen, topo, dut)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: routes are not present \n '
        'Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: routes are not present \n '
        'Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_add_and_remove_community_list_referenced_by_rmap_p0():
    """
    TC_51:
    Add and remove community-list referencec by route-map match cluase
    and verifying it reflecting as intended
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Creating configuration from JSON
    # build_config_from_json(tgen, topo)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_5 = {
        "r1": {
            "route_maps": {
                "rm_r1_out_{}".format(addr_type): [{
                    "action": "permit",
                    "set": {
                        "large_community": {"num": "1:1:1 1:2:3 2:1:1 2:2:2"}
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_5)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_6 = {
       'r1': {
           "bgp": {
               "address_family": {
                   "ipv4": {
                       "unicast": {
                           "neighbor": {
                               "r3": {
                                   "dest_link": {
                                       "r1": {
                                           "route_maps": [{
                                                   "name": "rm_r1_out_ipv4",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   },
                   "ipv6": {
                       "unicast": {
                           "neighbor": {
                               "r3": {
                                   "dest_link": {
                                       "r1": {
                                           "route_maps": [{
                                                   "name": "rm_r1_out_ipv6",
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
       }
   }

    result = create_router_bgp(tgen, topo, input_dict_6)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    for addr_type in ADDR_TYPES:
        # Create standard large commumity-list
            input_dict_1 = {
                "r3": {
                    "bgp_community_lists": [
                        {
                            "community_type": "standard",
                            "action": "permit",
                            "name": "rmap_lcomm_{}".format(addr_type),
                            "value": "1:1:1 1:2:3 2:1:1 2:2:2",
                            "large": True
                        }
                    ]
                }
            }
            result = create_bgp_community_lists(tgen, input_dict_1)
            assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
                tc_name, result)

    for addr_type in ADDR_TYPES:
    # Create route map
        input_dict_2 = {
        "r3": {
            "route_maps": {
                "rm_r3_in_{}".format(addr_type): [{
                    "action": "permit",
                    "match": {
                        addr_type : {
                            "large-community-list": {"id": "rmap_lcomm_"+
                            addr_type}
                    }
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_2)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_3 = {
       'r3': {
           "bgp": {
               "address_family": {
                   "ipv4": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name": "rm_r3_in_ipv4",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   },
                   "ipv6": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name": "rm_r3_in_ipv6",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
       }
   }
    result = create_router_bgp(tgen, topo, input_dict_3)

    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    sleep(5)
    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verify large-community-list
    dut = 'r3'
    networks = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    input_dict_4 = {
        'largeCommunity': '1:1:1 1:2:3 2:1:1 2:2:2'
    }
    for addr_type in ADDR_TYPES:
        result = verify_bgp_community(tgen, addr_type, dut, networks[
            addr_type],input_dict_4)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)
    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_multiple_match_statement_in_route_map_logical_ORed_p0():
    """
    TC_45:
    Test multiple match statements as part of a route-map's single
    sequence number. (Logical OR-ed of multiple match statements)
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Api call to advertise networks
    input_dict_nw1 = {
            'r1': {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "advertise_networks": [
                                    {"network": '10.0.30.1/32'}
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "advertise_networks": [
                                    {"network": '1::1/128'}
                                ]
                            }
                        }
                    }
                }
            }
        }

    result = create_router_bgp(tgen, topo, input_dict_nw1)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Api call to advertise networks
    input_dict_nw2 = {
            'r1': {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "advertise_networks": [
                                    {"network": '20.0.30.1/32'}
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "advertise_networks": [
                                    {"network": '2::1/128'}
                                ]
                            }
                        }
                    }
                }
            }
        }

    result = create_router_bgp(tgen, topo, input_dict_nw2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_2_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_2_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    input_dict_3_addr_type ={}
    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                    },
                    "set": {
                        "localpref": 150
                    }
                }]
            }
        }
        }
        input_dict_3_addr_type[addr_type] = input_dict_3
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "localpref": 200
                    }
                }]
            }
        }
        }
        input_dict_3_addr_type[addr_type] = input_dict_3
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_6 = {
       'r3': {
           "bgp": {
               "address_family": {
                   "ipv4": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv4",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   },
                   "ipv6": {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                                   "rmap_match_pf_1_ipv6",
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
       }
   }

    result = create_router_bgp(tgen, topo, input_dict_6)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.30.1/32"],
          "ipv6": ["1::1/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                rmap_name, input_dict_3_addr_type[addr_type])
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    routes = {
          "ipv4": ["20.0.30.1/32"],
          "ipv6": ["2::1/128"]
    }
    for addr_type in ADDR_TYPES:
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_multiple_match_statement_in_route_map_logical_ANDed():
    """
    TC_44:
    Test multiple match statements as part of a route-map's single
    sequence number. (Logical AND of multiple match statements)
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_5 = {
            "r1": {
                "route_maps": {
                    "rm_r1_out_{}".format(addr_type): [{
                        "action": "permit",
                        "set": {
                            "large_community": {
                                "num": "1:1:1 1:2:3 2:1:1 2:2:2"}
                        }
                    }]
                }
            }
        }
        result = create_route_maps(tgen, input_dict_5)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    for addr_type in ADDR_TYPES:
        input_dict_6 = {
        'r1': {
           "bgp": {
               "address_family": {
                addr_type: {
                       "unicast": {
                           "neighbor": {
                               "r3": {
                                   "dest_link": {
                                       "r1": {
                                           "route_maps": [{
                                                   "name":
                                            "rm_r1_out_{}".format(addr_type),
                                                   "direction": 'out'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
        }
        }
        result = create_router_bgp(tgen, topo, input_dict_6)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
                },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
                }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)

    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    for addr_type in ADDR_TYPES:
    # Create standard large commumity-list
        input_dict_1 = {
            "r3": {
                "bgp_community_lists": [
                    {
                        "community_type": "standard",
                        "action": "permit",
                        "name": "rmap_lcomm_{}".format(addr_type),
                        "value": "1:1:1 1:2:3 2:1:1 2:2:2",
                        "large": True
                    }
                ]
            }
        }
        result = create_bgp_community_lists(tgen, input_dict_1)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
            "r3": {
                "route_maps": {
                    "rmap_match_pf_1_{}".format(addr_type): [{
                        "action": "permit",
                        'seq_id': '5',
                        "match": {
                            addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                        },
                        "set": {
                            "localpref": 150,
                        }
                    }]
                }
            }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    for addr_type in ADDR_TYPES:
    # Create route map
        input_dict_3 = {
            "r3": {
                "route_maps": {
                    "rmap_match_pf_1_{}".format(addr_type): [{
                        "action": "permit",
                        'seq_id': '5',
                        "match": {
                            addr_type : {
                            "large_community_list": {"id": "rmap_lcomm_"+
                                addr_type}
                        }
                        },
                        "set": {
                            "localpref": 150,
                        }
                    }]
                }
            }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)
    # Configure neighbor for route map
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
       'r3': {
           "bgp": {
               "address_family": {
                addr_type: {
                       "unicast": {
                           "neighbor": {
                               "r1": {
                                   "dest_link": {
                                       "r3": {
                                           "route_maps": [{
                                                   "name":
                                        "rmap_match_pf_1_{}".format(addr_type),
                                                   "direction": 'in'
                                           }]
                                       }
                                   }
                               }
                           }
                       }
                   }
               }
           }
        }
        }
        result = create_router_bgp(tgen, topo, input_dict_4)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)
    # sleep(10)
    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_add_remove_rmap_to_specific_neighbor_p0():
    """
    TC_41:
    Test add/remove route-maps to specific neighbor and see if
    it takes effect as intended
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
            'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'deny'
                }]
            },
            'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'deny'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "localpref": 150,
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
        'r3': {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv4",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv6",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
         tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
                            protocol=protocol)
        assert result is not True, 'Testcase {} : \n'
        'Expected Behavior: Routes are not present in RIB \n'
        ' Error: {}'.format(
            tc_name, result)

    # Remove applied rmap from neighbor
    input_dict_4 = {
        'r3': {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv4",
                                                    "direction": 'in',
                                                    "delete": True
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv6",
                                                    "direction": 'in',
                                                    "delete": True
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
         tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
             tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_clear_bgp_and_flap_interface_to_verify_rmap_properties_p0():
    """
    TC_56:
    Test clear BGP sessions and interface flaps to see if
    route-map properties are intact.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5',
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "localpref": 150,
                        "weight": 100
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
        'r3': {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv4",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv6",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
         tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # clear bgp, so config changes would be reflected
    dut = 'r3'
    result = clear_bgp_and_verify(tgen, topo, dut)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                       rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Flap interface to see if route-map properties are intact
    # Shutdown interface
    dut = "r3"
    intf = "r3-r1-eth0"
    shutdown_bringup_interface(tgen, dut, intf, False)

    sleep(5)

    # Bringup interface
    dut = "r3"
    intf = "r3-r1-eth0"
    shutdown_bringup_interface(tgen, dut, intf, True)

    # Verify BGP convergence once interface is up
    result = verify_bgp_convergence(tgen, topo)
    assert result is True, (
        'setup_module :Failed \n Error:' ' {}'.format(result))

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_rmap_without_match_and_set_clause_p0():
    """
    TC_46:
    Verify if a blank sequence number can be create(without any
    match/set clause) and check if it allows all the traffic/prefixes
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_no_match_set_1_{}".format(addr_type): [{
                    "action": "permit",
                    'seq_id': '5'
                }],
                "rmap_no_match_set_2_{}".format(addr_type): [{
                    "action": "deny",
                    'seq_id': '5'
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
        'r3': {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_no_match_set_1_ipv4",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                },
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_no_match_set_2_ipv4",
                                                    "direction": 'out'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_no_match_set_1_ipv6",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                },
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_no_match_set_2_ipv6",
                                                    "direction": 'out'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
         tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: routes are not present \n '
        'Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_set_localpref_weight_to_ebgp_and_med_to_ibgp_peers_p0():
    """
    TC_48:
    Create route map setting local preference and weight to eBGP peeer
    and metric to ibgp peer and verifying it should not get advertised
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                  'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                  'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    input_dict_3_addr_type ={}
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
            "r3": {
                "route_maps": {
                    "rmap_match_pf_1_{}".format(addr_type): [{
                            "action": "permit",
                            "match": {
                                addr_type: {
                                    "prefix_lists": "pf_list_1_{}".format(
                                        addr_type)
                                }
                            },
                                "set": {
                                    "med": 50
                                }
                            }],
                    "rmap_match_pf_2_{}".format(addr_type): [{
                            "action": "permit",
                            "match": {
                                addr_type: {
                                    "prefix_lists": "pf_list_1_{}".format(
                                            addr_type)
                                    }},
                                    "set": {
                                        "localpref": 150
                                    }
                                }],
                            "rmap_match_pf_3_{}".format(addr_type): [{
                                    "action": "permit",
                                    "match": {
                                        addr_type: {
                                    "prefix_lists": "pf_list_1_{}".format(
                                            addr_type)
                                        }},
                                        "set": {
                                            "weight": 1000
                                        }
                                    }]
                            }
                        }
                    }
        input_dict_3_addr_type[addr_type] = input_dict_3
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
        'r3': {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv4",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                },
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_2_ipv4",
                                                    "direction": 'out'
                                            }]
                                        }
                                    }
                                },
                                "r5": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_3_ipv4",
                                                    "direction": 'out'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_1_ipv6",
                                                    "direction": 'in'
                                            }]
                                        }
                                    }
                                },
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_2_ipv6",
                                                    "direction": 'out'
                                            }]
                                        }
                                    }
                                },
                                "r5": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [{
                                                    "name":
                                                    "rmap_match_pf_3_ipv6",
                                                    "direction": 'out'
                                            }]
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    rmap_name = "rmap_match_pf_1"
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[
            addr_type],rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r4'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    rmap_name = "rmap_match_pf_2"
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_2_{}".format(addr_type)

        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                rmap_name, input_dict_3_addr_type[addr_type])
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: Attributes are not set \n'
        'Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r5'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)
    # Verifying BGP set attributes
    dut = 'r5'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }

    rmap_name = "rmap_match_pf_3"
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_3_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                rmap_name, input_dict_3_addr_type[addr_type])
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: Attributes are not set \n'
        'Error: {}'.format(
            tc_name, result)

        logger.info("Expected behaviour: {}".format(result))

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_multiple_set_on_single_sequence_in_rmap_p0():
    """
    TC_43:
    Test multiple set statements as part of a route-map's
    single sequence number.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
                },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
                }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                },
                    "set": {
                        "localpref": 150,
                        "weight": 100,
                        "med": 50
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv4",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                     "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv6",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
     }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }

    rmap_name = "rmap_match_pf_1"
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_route_maps_with_continue_clause_p0():
    """
    TC_54:
    Verify route-maps continue clause functionality.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                        "action": "permit",
                        'seq_id': '10',
                        "match": {
                            addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                        },
                        "set": {
                            "localpref": 150
                        },
                        "continue": "30"
                    },
                    {
                        "action": "permit",
                        'seq_id': '20',
                        "match": {
                            addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                        },
                        "set": {
                            "med": 200
                        }
                    },
                    {
                        "action": "permit",
                        'seq_id': '30',
                        "match": {
                            addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                        },
                        "set": {
                            "med": 100
                        }
                    }
                ]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv4",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                     "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv6",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
     }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
            protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    rmap_name = "rmap_match_pf_1"
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    seq_id = {
          "ipv4": ["10", "30"],
          "ipv6": ["10", "30"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[
            addr_type],rmap_name, input_dict_3, seq_id[addr_type])
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_route_maps_with_goto_clause_p0():
    """
    TC_55:
    Verify route-maps goto clause functionality.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                        "action": "permit",
                        'seq_id': '10',
                        "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                        },
                        "goto": "30"
                    },
                    {
                        "action": "permit",
                        'seq_id': '20',
                        "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                        },
                        "set": {
                            "med": 100
                        }
                    },
                    {
                        "action": "permit",
                        'seq_id': '30',
                        "match": {
                        addr_type: {
                            "prefix_lists": "pf_list_1_{}".format(addr_type)
                        }
                        },
                        "set": {
                            "med": 200
                        }
                    }
                ]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv4",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                     "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv6",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
     }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    rmap_name = "rmap_match_pf_1"
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    seq_id = {
          "ipv4": ["10", "30"],
          "ipv6": ["10", "30"]
    }
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[
            addr_type],rmap_name, input_dict_3, seq_id[addr_type])
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_route_maps_with_call_clause_p0():
    """
    TC_53:
    Verify route-maps call clause functionality.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "permit",
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "localpref": 150
                    },
                    "call": "rmap_match_pf_2_{}".format(addr_type)
                }],
                "rmap_match_pf_2_{}".format(addr_type): [{
                    "action": "permit",
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "med": 200
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv4",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                     "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv6",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
     }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Verifying BGP set attributes
    dut = 'r3'
    routes = {
          "ipv4": ["10.0.20.1/32", "10.0.20.2/32"],
          "ipv6": ["1::1/128", "1::2/128"]
    }
    rmap_name = "rmap_match_pf_1"
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_1_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                   rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    rmap_name = "rmap_match_pf_2"
    for addr_type in ADDR_TYPES:
        rmap_name = "rmap_match_pf_2_{}".format(addr_type)
        result = verify_bgp_attributes(tgen, addr_type, dut, routes[addr_type],
                                    rmap_name, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_create_rmap_match_prefix_list_to_deny_in_and_outbound_prefixes_p0():
    """
    TC_58:
    Create route map deny inbound and outbound prefixes on
    match prefix list and set criteria on match
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    # Create ip prefix list
    input_dict_2 = {
        'r3': {
            'prefix_lists': {
                'ipv4': {
                'pf_list_1_ipv4': [{
                    'seqid': 10,
                    'network': 'any',
                    'action': 'permit'
                }]
            },
                'ipv6': {
                'pf_list_1_ipv6': [{
                    'seqid': 100,
                    'network': 'any',
                    'action': 'permit'
                }]
            }
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Create route map
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
        "r3": {
            "route_maps": {
                "rmap_match_pf_1_{}".format(addr_type): [{
                    "action": "deny",
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "localpref": 150,
                    }
                }],
                "rmap_match_pf_2_{}".format(addr_type): [{
                    "action": "deny",
                    "match": {
                        addr_type: {
                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                    }
                    },
                    "set": {
                        "med": 50
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r3': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv4",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 },
                                 "r4": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_2_ipv6",
                                                     "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                     "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r1": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_1_ipv4",
                                                     "direction": 'in'
                                             }]
                                         }
                                     }
                                 },
                                 "r4": {
                                     "dest_link": {
                                         "r3": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_pf_2_ipv6",
                                                     "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
     }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'
    input_dict = topo["routers"]
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
        protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: routes are not present \n '
        'Error: {}'.format(
            tc_name, result)

    # Verifying RIB routes
    dut = 'r4'
    protocol = 'bgp'
    for addr_type in ADDR_TYPES:
        result = verify_rib(tgen, addr_type, dut, input_dict,
        protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behaviour: routes are not present \n '
        'Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_create_rmap_to_match_tag_permit_inbound_prefixes_p0():
    """
    TC_59:
    Create route map to permit inbound prefixes with filter
    match tag and set criteria
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    for addr_type in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "tag": 4001
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result)

        # Api call to redistribute static routes
        input_dict_1 = {
            "r1": {
                "bgp": {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"}
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"}
                                ]
                            }
                        }
                    }
                }
            }
        }

        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result)

        # Create route map
        input_dict_3 = {
        "r1": {
            "route_maps": {
                "rmap_match_tag_1_{}".format(addr_type): [{
                    "action": "permit",
                    "match": {
                        addr_type: {
                        "tag": "4001"
                    }
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r1': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r3": {
                                     "dest_link": {
                                         "r1": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_tag_1_ipv4",
                                                     "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                     "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r3": {
                                     "dest_link": {
                                         "r1": {
                                             "route_maps": [{
                                                     "name":
                                                     "rmap_match_tag_1_ipv6",
                                                     "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
     }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "tag": 4001
                    }
                ]
            }
        }
        result = verify_rib(tgen, addr_type, dut, input_dict,
        protocol=protocol)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()


def test_create_rmap_to_match_tag_deny_outbound_prefixes_p0():
    """
    TC_60
    Create route map to deny outbound prefixes with filter match tag,
    and set criteria
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip('skipped because of BGP Convergence failure')

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    reset_config_on_routers(tgen)

    for addr_type in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "tag": 4001
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result)

        # Api call to redistribute static routes
        input_dict_1 = {
            "r1": {
                "bgp": {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"}
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"}
                                ]
                            }
                        }
                    }
                }
            }
        }

        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result)

        # Create route map
        input_dict_3 = {
        "r1": {
            "route_maps": {
                "rmap_match_tag_1_{}".format(addr_type): [{
                    "action": "deny",
                    "match": {
                        addr_type: {
                        "tag": "4001"
                    }
                    }
                }]
            }
        }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
            tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
       'r1': {
             "bgp": {
                 "address_family": {
                     "ipv4": {
                         "unicast": {
                             "neighbor": {
                                 "r3": {
                                     "dest_link": {
                                         "r1": {
                                             "route_maps": [{
                                                    "name":
                                                    "rmap_match_tag_1_ipv4",
                                                    "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     },
                     "ipv6": {
                         "unicast": {
                             "neighbor": {
                                 "r3": {
                                     "dest_link": {
                                         "r1": {
                                             "route_maps": [{
                                                    "name":
                                                    "rmap_match_tag_1_ipv6",
                                                    "direction": 'out'
                                             }]
                                         }
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
         }
    }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, 'Testcase {} : Failed \n Error: {}'.format(
        tc_name, result)

    # Verifying RIB routes
    dut = 'r3'
    protocol = 'bgp'

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "tag": 4001
                    }
                ]
            }
        }
        result = verify_rib(tgen, addr_type, dut, input_dict,
                            protocol=protocol)
        assert result is not True, 'Testcase {} : Failed \n'
        'Expected behavior: routes are denied \n Error: {}'.format(
            tc_name, result)

    write_test_footer(tc_name)

    # Uncomment next line for debugging
    # tgen.mininet_cli()

if __name__ == '__main__':
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python

#
# test_bfd_bgp_cbit_topo3.py
#
# Copyright (c) 2019 6WIND
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NETDEF DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NETDEF BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#

"""
test_bfd_bgp_cbit_topo3.py: Test the FRR/Quagga BFD daemon with multihop and BGP
unnumbered.
"""

import os
import sys
import json
from functools import partial
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, '../'))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.
from mininet.topo import Topo

class BFDTopo(Topo):
    "Test topology builder"
    def build(self, *_args, **_opts):
        "Build function"
        tgen = get_topogen(self)

        # Create 4 routers.
        for routern in range(1, 4):
            tgen.add_router('r{}'.format(routern))

        switch = tgen.add_switch('s1')
        switch.add_link(tgen.gears['r1'])
        switch.add_link(tgen.gears['r2'])

        switch = tgen.add_switch('s2')
        switch.add_link(tgen.gears['r2'])
        switch.add_link(tgen.gears['r3'])

def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(BFDTopo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for rname, router in router_list.iteritems():
        router.load_config(
            TopoRouter.RD_ZEBRA,
            os.path.join(CWD, '{}/zebra.conf'.format(rname)),
        )
        router.load_config(
            TopoRouter.RD_BFD,
            os.path.join(CWD, '{}/bfdd.conf'.format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP,
            os.path.join(CWD, '{}/bgpd.conf'.format(rname))
        )

    # Initialize all routers.
    tgen.start_router()

    # Verify that we are using the proper version and that the BFD
    # daemon exists.
    for router in router_list.values():
        # Check for Version
        if router.has_version('<', '5.1'):
            tgen.set_error('Unsupported FRR version')
            break

def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


def test_protocols_convergence():
    """
    Assert that all protocols have converged before checking for the BFD
    statuses as they depend on it.
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Check IPv6 routing tables.
    logger.info("Checking IPv6 routes for convergence")
    for router in tgen.routers().values():
        if router.name == 'r2':
            continue
        json_file = '{}/{}/ipv6_routes.json'.format(CWD, router.name)
        if not os.path.isfile(json_file):
            logger.info('skipping file {}'.format(json_file))
            continue
        expected = json.loads(open(json_file).read())
        test_func = partial(topotest.router_json_cmp,
                            router, 'show ipv6 route json', expected)
        _, result = topotest.run_and_expect(test_func, None, count=40,
                                            wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg


def test_bfd_connection():
    "Assert that the BFD peers can find themselves."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info('waiting for bfd peers to go up')
    for router in tgen.routers().values():
        if router.name == 'r2':
            continue
        json_file = '{}/{}/peers.json'.format(CWD, router.name)
        expected = json.loads(open(json_file).read())

        test_func = partial(topotest.router_json_cmp,
                            router, 'show bfd peers json', expected)
        _, result = topotest.run_and_expect(test_func, None, count=8, wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg

def test_bfd_loss_intermediate():
    """
    Assert that BFD notices the bfd link down failure.
    but BGP entries should still be present
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info('removing IPv6 address from r2 to simulate loss of connectivity')
    # Disable r2-eth0 ipv6 address
    cmd = 'vtysh -c \"configure terminal\" -c \"interface r2-eth1\" -c "no ipv6 address 2001:db8:4::2/64\"'
    tgen.net['r2'].cmd(cmd)
 
    # Wait the minimum time we can before checking that BGP/BFD
    # converged.
    logger.info('waiting for BFD converge down')

    # Check that BGP converged quickly.
    for router in tgen.routers().values():
        if router.name == 'r2':
            continue
        json_file = '{}/{}/peers_down.json'.format(CWD, router.name)
        expected = json.loads(open(json_file).read())

        test_func = partial(topotest.router_json_cmp,
                            router, 'show bfd peers json', expected)
        _, result = topotest.run_and_expect(test_func, None, count=8, wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg

    logger.info('waiting for BGP entries to become stale')
    for router in tgen.routers().values():
        if router.name == 'r2':
            continue
        json_file = '{}/{}/bgp_ipv6_routes_down.json'.format(CWD, router.name)
        expected = json.loads(open(json_file).read())

        test_func = partial(topotest.router_json_cmp,
                            router, 'show bgp ipv6 json', expected)
        _, result = topotest.run_and_expect(test_func, None, count=50, wait=1)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg

    logger.info("Checking IPv6 routes on r1 should still be present")
    for router in tgen.routers().values():
        if router.name == 'r2':
            continue
        if router.name == 'r3':
            continue
        json_file = '{}/r1/ipv6_routes.json'.format(CWD)
        expected = json.loads(open(json_file).read())
        test_func = partial(topotest.router_json_cmp,
                            router, 'show ipv6 route json', expected)
        _, result = topotest.run_and_expect(test_func, None, count=30,
                                            wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg

def test_bfd_comes_back_again():
    """
    Assert that BFD notices the bfd link up
    and that ipv6 entries appear back
    """
    tgen = get_topogen()
    logger.info('re-adding IPv6 address from r2 to simulate connectivity is back')
    # adds back r2-eth0 ipv6 address
    cmd = 'vtysh -c \"configure terminal\" -c \"interface r2-eth1\" -c "ipv6 address 2001:db8:4::2/64\"'
    tgen.net['r2'].cmd(cmd)

    # Wait the minimum time we can before checking that BGP/BFD
    # converged.
    logger.info('waiting for BFD to converge up')

    # Check that BGP converged quickly.
    for router in tgen.routers().values():
        if router.name == 'r2':
            continue
        json_file = '{}/{}/peers.json'.format(CWD, router.name)
        expected = json.loads(open(json_file).read())

        test_func = partial(topotest.router_json_cmp,
                            router, 'show bfd peers json', expected)
        _, result = topotest.run_and_expect(test_func, None, count=8, wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg
    

def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip('Memory leak test/report is disabled')

    tgen.report_memory_leaks()


if __name__ == '__main__':
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



from lutil import luCommand
holddownFactorSet = luCommand('r1','vtysh -c "show running"','rfp holddown-factor','none','Holddown factor set')
if not holddownFactorSet:
    to = "-1"
    cost = ""
else:
    to = "6"
    cost = "cost 50"
luCommand('r1','vtysh -c "debug rfapi-dev open vn 10.0.0.1 un 1.1.1.1"','rfapi_set_response_cb: status 0', 'pass', 'Opened RFAPI')
luCommand('r1','vtysh -c "debug rfapi-dev query vn 10.0.0.1 un 1.1.1.1 target 11.11.11.11"','rc=2', 'pass', 'Clean query')
luCommand('r1','vtysh -c "debug rfapi-dev register vn 10.0.0.1 un 1.1.1.1 prefix 11.11.11.0/24 lifetime {}"'.format(to),'', 'none', 'Prefix registered')
luCommand('r1','vtysh -c "show vnc registrations local"','1 out of 1','wait','Local registration')
luCommand('r1','vtysh -c "debug rfapi-dev response-omit-self off"','.','none')
luCommand('r1','vtysh -c "debug rfapi-dev query vn 10.0.0.1 un 1.1.1.1 target 11.11.11.11"','11.11.11.0/24', 'pass', 'Query self')

luCommand('r3','vtysh -c "debug rfapi-dev open vn 10.0.0.2 un 2.2.2.2"','rfapi_set_response_cb: status 0', 'pass', 'Opened RFAPI')
luCommand('r3','vtysh -c "debug rfapi-dev register vn 10.0.0.2 un 2.2.2.2 prefix 22.22.22.0/24 lifetime {}"'.format(to),'', 'none', 'Prefix registered')
luCommand('r3','vtysh -c "show vnc registrations local"','1 out of 1','wait','Local registration')
luCommand('r3','vtysh -c "debug rfapi-dev response-omit-self on"','.','none')
luCommand('r3','vtysh -c "debug rfapi-dev query vn 10.0.0.2 un 2.2.2.2 target 22.22.22.22"','rc=2', 'pass', 'Self excluded')
luCommand('r3','vtysh -c "debug rfapi-dev open vn 10.0.1.2 un 2.1.1.2"','rfapi_set_response_cb: status 0', 'pass', 'Opened query only RFAPI')
luCommand('r3','vtysh -c "debug rfapi-dev query vn 10.0.1.2 un 2.1.1.2 target 22.22.22.22"','22.22.22.0/24', 'pass', 'See local')

luCommand('r4','vtysh -c "debug rfapi-dev open vn 10.0.0.3 un 3.3.3.3"','rfapi_set_response_cb: status 0', 'pass', 'Opened RFAPI')
luCommand('r4','vtysh -c "debug rfapi-dev register vn 10.0.0.3 un 3.3.3.3 prefix 33.33.33.0/24 lifetime {}"'.format(to),'', 'none', 'Prefix registered')
luCommand('r4','vtysh -c "show vnc registrations local"','1 out of 1','wait','Local registration')
luCommand('r4','vtysh -c "debug rfapi-dev response-omit-self off"','.','none')
luCommand('r4','vtysh -c "debug rfapi-dev query vn 10.0.0.3 un 3.3.3.3 target 33.33.33.33"','33.33.33.0/24', 'pass', 'Query self')

luCommand('r4','vtysh -c "debug rfapi-dev register vn 10.0.0.3 un 3.3.3.3 prefix 11.11.11.0/24 lifetime {} {}"'.format(to, cost),'', 'none', 'MP Prefix registered')
luCommand('r4','vtysh -c "show vnc registrations local"','2 out of 2','wait','Local registration')
luCommand('r4','vtysh -c "debug rfapi-dev query vn 10.0.0.3 un 3.3.3.3 target 11.11.11.11"','11.11.11.0/24', 'pass', 'Query self MP')

luCommand('r1','vtysh -c "show vnc registrations"','.','none')
luCommand('r3','vtysh -c "show vnc registrations"','.','none')
luCommand('r4','vtysh -c "show vnc registrations"','.','none')



#!/usr/bin/env python

#
# test_ospf_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NETDEF DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NETDEF BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#

"""
test_ospf_topo1.py: Test the FRR/Quagga OSPF routing daemon.
"""

import os
import re
import sys
from functools import partial
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, '../'))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.
from mininet.topo import Topo

class OSPFTopo(Topo):
    "Test topology builder"
    def build(self, *_args, **_opts):
        "Build function"
        tgen = get_topogen(self)

        # Create 4 routers
        for routern in range(1, 5):
            tgen.add_router('r{}'.format(routern))

        # Create a empty network for router 1
        switch = tgen.add_switch('s1')
        switch.add_link(tgen.gears['r1'])

        # Create a empty network for router 2
        switch = tgen.add_switch('s2')
        switch.add_link(tgen.gears['r2'])

        # Interconect router 1, 2 and 3
        switch = tgen.add_switch('s3')
        switch.add_link(tgen.gears['r1'])
        switch.add_link(tgen.gears['r2'])
        switch.add_link(tgen.gears['r3'])

        # Create empty netowrk for router3
        switch = tgen.add_switch('s4')
        switch.add_link(tgen.gears['r3'])

        # Interconect router 3 and 4
        switch = tgen.add_switch('s5')
        switch.add_link(tgen.gears['r3'])
        switch.add_link(tgen.gears['r4'])

        # Create a empty network for router 4
        switch = tgen.add_switch('s6')
        switch.add_link(tgen.gears['r4'])

def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(OSPFTopo, mod.__name__)
    tgen.start_topology()

    ospf6_config = 'ospf6d.conf'
    if tgen.gears['r1'].has_version('<', '4.0'):
        ospf6_config = 'ospf6d.conf-pre-v4'

    router_list = tgen.routers()
    for rname, router in router_list.iteritems():
        router.load_config(
            TopoRouter.RD_ZEBRA,
            os.path.join(CWD, '{}/zebra.conf'.format(rname))
        )
        router.load_config(
            TopoRouter.RD_OSPF,
            os.path.join(CWD, '{}/ospfd.conf'.format(rname))
        )
        router.load_config(
            TopoRouter.RD_OSPF6,
            os.path.join(CWD, '{}/{}'.format(rname, ospf6_config))
        )

    # Initialize all routers.
    tgen.start_router()

def teardown_module(mod):
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


def compare_show_ipv6_ospf6(rname, expected):
    """
    Calls 'show ipv6 ospf6 route' for router `rname` and compare the obtained
    result with the expected output.
    """
    tgen = get_topogen()
    current = tgen.gears[rname].vtysh_cmd('show ipv6 ospf6 route')

    # Remove the link addresses
    current = re.sub(r'fe80::[^ ]+', 'fe80::xxxx:xxxx:xxxx:xxxx', current)
    expected = re.sub(r'fe80::[^ ]+', 'fe80::xxxx:xxxx:xxxx:xxxx', expected)

    # Remove the time
    current = re.sub(r'\d+:\d{2}:\d{2}', '', current)
    expected = re.sub(r'\d+:\d{2}:\d{2}', '', expected)

    return topotest.difflines(topotest.normalize_text(current),
                              topotest.normalize_text(expected),
                              title1="Current output",
                              title2="Expected output")

def test_ospf_convergence():
    "Test OSPF daemon convergence"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    for router, rnode in tgen.routers().iteritems():
        logger.info('Waiting for router "%s" convergence', router)

        # Load expected results from the command
        reffile = os.path.join(CWD, '{}/ospfroute.txt'.format(router))
        expected = open(reffile).read()

        # Run test function until we get an result. Wait at most 80 seconds.
        test_func = partial(
            topotest.router_output_cmp, rnode, 'show ip ospf route', expected)
        result, diff = topotest.run_and_expect(test_func, '',
                                               count=160, wait=0.5)
        assert result, 'OSPF did not converge on {}:\n{}'.format(router, diff)

def test_ospf_kernel_route():
    "Test OSPF kernel route installation"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    rlist = tgen.routers().values()
    for router in rlist:
        logger.info('Checking OSPF IPv4 kernel routes in "%s"', router.name)

        routes = topotest.ip4_route(router)
        expected = {
            '10.0.1.0/24': {},
            '10.0.2.0/24': {},
            '10.0.3.0/24': {},
            '10.0.10.0/24': {},
            '172.16.0.0/24': {},
            '172.16.1.0/24': {},
        }
        assertmsg = 'OSPF IPv4 route mismatch in router "{}"'.format(router.name)
        assert topotest.json_cmp(routes, expected) is None, assertmsg

def test_ospf6_convergence():
    "Test OSPF6 daemon convergence"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    ospf6route_file = '{}/ospf6route_ecmp.txt'
    for rnum in range(1, 5):
        router = 'r{}'.format(rnum)

        logger.info('Waiting for router "%s" IPv6 OSPF convergence', router)

        # Load expected results from the command
        reffile = os.path.join(CWD, ospf6route_file.format(router))
        expected = open(reffile).read()

        # Run test function until we get an result. Wait at most 60 seconds.
        test_func = partial(compare_show_ipv6_ospf6, router, expected)
        result, diff = topotest.run_and_expect(test_func, '',
                                               count=25, wait=3)
        if (not result) and (rnum == 1):
            # Didn't match the new ECMP version - try the old pre-ECMP format
            ospf6route_file = '{}/ospf6route.txt'

            # Load expected results from the command
            reffile = os.path.join(CWD, ospf6route_file.format(router))
            expected = open(reffile).read()

            test_func = partial(compare_show_ipv6_ospf6, router, expected)
            result, diff = topotest.run_and_expect(test_func, '',
                                               count=1, wait=3)
            if not result:
                # Didn't match the old version - switch back to new ECMP version
                # and fail
                ospf6route_file = '{}/ospf6route_ecmp.txt'

                # Load expected results from the command
                reffile = os.path.join(CWD, ospf6route_file.format(router))
                expected = open(reffile).read()

                test_func = partial(compare_show_ipv6_ospf6, router, expected)
                result, diff = topotest.run_and_expect(test_func, '',
                                               count=1, wait=3)

        assert result, 'OSPF6 did not converge on {}:\n{}'.format(router, diff)

def test_ospf6_kernel_route():
    "Test OSPF kernel route installation"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    rlist = tgen.routers().values()
    for router in rlist:
        logger.info('Checking OSPF IPv6 kernel routes in "%s"', router.name)

        routes = topotest.ip6_route(router)
        expected = {
            '2001:db8:1::/64': {},
            '2001:db8:2::/64': {},
            '2001:db8:3::/64': {},
            '2001:db8:100::/64': {},
            '2001:db8:200::/64': {},
            '2001:db8:300::/64': {},
        }
        assertmsg = 'OSPF IPv6 route mismatch in router "{}"'.format(router.name)
        assert topotest.json_cmp(routes, expected) is None, assertmsg

def test_ospf_json():
    "Test 'show ip ospf json' output for coherency."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    for rnum in range(1, 5):
        router = tgen.gears['r{}'.format(rnum)]
        logger.info('Comparing router "%s" "show ip ospf json" output', router.name)
        expected = {
            'routerId': '10.0.255.{}'.format(rnum),
            'tosRoutesOnly': True,
            'rfc2328Conform': True,
            'spfScheduleDelayMsecs': 0,
            'holdtimeMinMsecs': 50,
            'holdtimeMaxMsecs': 5000,
            'lsaMinIntervalMsecs': 5000,
            'lsaMinArrivalMsecs': 1000,
            'writeMultiplier': 20,
            'refreshTimerMsecs': 10000,
            'asbrRouter': 'injectingExternalRoutingInformation',
            'attachedAreaCounter': 1,
            'areas': {}
        }
        # Area specific additional checks
        if router.name == 'r1' or router.name == 'r2' or router.name == 'r3':
            expected['areas']['0.0.0.0'] = {
                'areaIfActiveCounter': 2,
                'areaIfTotalCounter': 2,
                'authentication': 'authenticationNone',
                'backbone': True,
                'lsaAsbrNumber': 1,
                'lsaNetworkNumber': 1,
                'lsaNssaNumber': 0,
                'lsaNumber': 7,
                'lsaOpaqueAreaNumber': 0,
                'lsaOpaqueLinkNumber': 0,
                'lsaRouterNumber': 3,
                'lsaSummaryNumber': 2,
                'nbrFullAdjacentCounter': 2,
            }
        if router.name == 'r3' or router.name == 'r4':
            expected['areas']['0.0.0.1'] = {
                'areaIfActiveCounter': 1,
                'areaIfTotalCounter': 1,
                'authentication': 'authenticationNone',
                'lsaAsbrNumber': 2,
                'lsaNetworkNumber': 1,
                'lsaNssaNumber': 0,
                'lsaNumber': 9,
                'lsaOpaqueAreaNumber': 0,
                'lsaOpaqueLinkNumber': 0,
                'lsaRouterNumber': 2,
                'lsaSummaryNumber': 4,
                'nbrFullAdjacentCounter': 1,
            }
            # r4 has more interfaces for area 0.0.0.1
            if router.name == 'r4':
                expected['areas']['0.0.0.1'].update({
                    'areaIfActiveCounter': 2,
                    'areaIfTotalCounter': 2,
                })

        # router 3 has an additional area
        if router.name == 'r3':
            expected['attachedAreaCounter'] = 2

        output = router.vtysh_cmd('show ip ospf json', isjson=True)
        result = topotest.json_cmp(output, expected)
        assert result is None, '"{}" JSON output mismatches the expected result'.format(router.name)

def test_ospf_link_down():
    "Test OSPF convergence after a link goes down"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    # Simulate a network down event on router3 switch3 interface.
    router3 = tgen.gears['r3']
    router3.peer_link_enable('r3-eth0', False)

    # Expect convergence on all routers
    for router, rnode in tgen.routers().iteritems():
        logger.info('Waiting for router "%s" convergence after link failure', router)
        # Load expected results from the command
        reffile = os.path.join(CWD, '{}/ospfroute_down.txt'.format(router))
        expected = open(reffile).read()

        # Run test function until we get an result. Wait at most 80 seconds.
        test_func = partial(
            topotest.router_output_cmp, rnode, 'show ip ospf route', expected)
        result, diff = topotest.run_and_expect(test_func, '',
                                               count=140, wait=0.5)
        assert result, 'OSPF did not converge on {}:\n{}'.format(router, diff)

def test_ospf_link_down_kernel_route():
    "Test OSPF kernel route installation"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    rlist = tgen.routers().values()
    for router in rlist:
        logger.info('Checking OSPF IPv4 kernel routes in "%s" after link down', router.name)

        routes = topotest.ip4_route(router)
        expected = {
            '10.0.1.0/24': {},
            '10.0.2.0/24': {},
            '10.0.3.0/24': {},
            '10.0.10.0/24': {},
            '172.16.0.0/24': {},
            '172.16.1.0/24': {},
        }
        if router.name == 'r1' or router.name == 'r2':
            expected.update({
                '10.0.10.0/24': None,
                '172.16.0.0/24': None,
                '172.16.1.0/24': None,
            })
        elif router.name == 'r3' or router.name == 'r4':
            expected.update({
                '10.0.1.0/24': None,
                '10.0.2.0/24': None,
            })
        # Route '10.0.3.0' is no longer available for r4 since it is down.
        if router.name == 'r4':
            expected.update({
                '10.0.3.0/24': None,
            })
        assertmsg = 'OSPF IPv4 route mismatch in router "{}" after link down'.format(router.name)
        assert topotest.json_cmp(routes, expected) is None, assertmsg

def test_ospf6_link_down():
    "Test OSPF6 daemon convergence after link goes down"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    for rnum in range(1, 5):
        router = 'r{}'.format(rnum)

        logger.info('Waiting for router "%s" IPv6 OSPF convergence after link down', router)

        # Load expected results from the command
        reffile = os.path.join(CWD, '{}/ospf6route_down.txt'.format(router))
        expected = open(reffile).read()

        # Run test function until we get an result. Wait at most 60 seconds.
        test_func = partial(compare_show_ipv6_ospf6, router, expected)
        result, diff = topotest.run_and_expect(test_func, '',
                                               count=25, wait=3)
        assert result, 'OSPF6 did not converge on {}:\n{}'.format(router, diff)

def test_ospf6_link_down_kernel_route():
    "Test OSPF kernel route installation"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip('skipped because of router(s) failure')

    rlist = tgen.routers().values()
    for router in rlist:
        logger.info('Checking OSPF IPv6 kernel routes in "%s" after link down', router.name)

        routes = topotest.ip6_route(router)
        expected = {
            '2001:db8:1::/64': {},
            '2001:db8:2::/64': {},
            '2001:db8:3::/64': {},
            '2001:db8:100::/64': {},
            '2001:db8:200::/64': {},
            '2001:db8:300::/64': {},
        }
        if router.name == 'r1' or router.name == 'r2':
            expected.update({
                '2001:db8:100::/64': None,
                '2001:db8:200::/64': None,
                '2001:db8:300::/64': None,
            })
        elif router.name == 'r3' or router.name == 'r4':
            expected.update({
                '2001:db8:1::/64': None,
                '2001:db8:2::/64': None,
            })
        # Route '2001:db8:3::/64' is no longer available for r4 since it is down.
        if router.name == 'r4':
            expected.update({
                '2001:db8:3::/64': None,
            })
        assertmsg = 'OSPF IPv6 route mismatch in router "{}" after link down'.format(router.name)
        assert topotest.json_cmp(routes, expected) is None, assertmsg

def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip('Memory leak test/report is disabled')

    tgen.report_memory_leaks()

if __name__ == '__main__':
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python

#
# test_bgp_ipv6_rtadv.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2019 by 6WIND
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NETDEF DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NETDEF BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#

"""
 test_bgp_ipv6_rtadv.py: Test the FRR/Quagga BGP daemon with BGP IPv6 interface
 with route advertisements on a separate netns.
"""

import os
import sys
import json
from functools import partial
import pytest
import platform

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, '../'))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.
from mininet.topo import Topo


class BGPIPV6RTADVVRFTopo(Topo):
    "Test topology builder"
    def build(self, *_args, **_opts):
        "Build function"
        tgen = get_topogen(self)

        # Create 2 routers.
        tgen.add_router('r1')
        tgen.add_router('r2')

        switch = tgen.add_switch('s1')
        switch.add_link(tgen.gears['r1'])
        switch.add_link(tgen.gears['r2'])

def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(BGPIPV6RTADVVRFTopo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    logger.info('Testing with VRF Lite support')
    krel = platform.release()

    # May need to adjust handling of vrf traffic depending on kernel version
    l3mdev_accept = 0
    if topotest.version_cmp(krel, '4.15') >= 0 and \
       topotest.version_cmp(krel, '4.18') <= 0:
        l3mdev_accept = 1

    if topotest.version_cmp(krel, '5.0') >= 0:
        l3mdev_accept = 1

    logger.info('krel \'{0}\' setting net.ipv4.tcp_l3mdev_accept={1}'.format(
        krel, l3mdev_accept))

    cmds = ['ip link add {0}-cust1 type vrf table 1001',
            'ip link add loop1 type dummy',
            'ip link set loop1 master {0}-cust1',
            'ip link set {0}-eth0 master {0}-cust1']

    for rname, router in router_list.iteritems():
        for cmd in cmds:
            output = tgen.net[rname].cmd(cmd.format(rname))

        output = tgen.net[rname].cmd('sysctl -n net.ipv4.tcp_l3mdev_accept')
        logger.info(
            'router {0}: existing tcp_l3mdev_accept was {1}'.format(
                rname, output))

        if l3mdev_accept:
            output = tgen.net[rname].cmd(
                'sysctl -w net.ipv4.tcp_l3mdev_accept={}'.format(l3mdev_accept))

    for rname, router in router_list.iteritems():
        router.load_config(
            TopoRouter.RD_ZEBRA,
            os.path.join(CWD, '{}/zebra.conf'.format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP,
            os.path.join(CWD, '{}/bgpd.conf'.format(rname))
        )

    # Initialize all routers.
    tgen.start_router()

def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()

    tgen.stop_topology()


def test_protocols_convergence():
    """
    Assert that all protocols have converged
    statuses as they depend on it.
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Check IPv4 routing tables.
    logger.info("Checking IPv4 routes for convergence")

    for router in tgen.routers().values():
        json_file = '{}/{}/ipv4_routes.json'.format(CWD, router.name)
        if not os.path.isfile(json_file):
            logger.info('skipping file {}'.format(json_file))
            continue

        expected = json.loads(open(json_file).read())
        test_func = partial(topotest.router_json_cmp,
                            router, 'show ip route vrf {}-cust1 json'.format(router.name), expected)
        _, result = topotest.run_and_expect(test_func, None, count=160,
                                            wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg

    # Check IPv6 routing tables.
    logger.info("Checking IPv6 routes for convergence")
    for router in tgen.routers().values():
        json_file = '{}/{}/ipv6_routes.json'.format(CWD, router.name)
        if not os.path.isfile(json_file):
            logger.info('skipping file {}'.format(json_file))
            continue

        expected = json.loads(open(json_file).read())
        test_func = partial(topotest.router_json_cmp,
                            router, 'show ipv6 route vrf {}-cust1 json'.format(router.name), expected)
        _, result = topotest.run_and_expect(test_func, None, count=160,
                                            wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg

def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip('Memory leak test/report is disabled')

    tgen.report_memory_leaks()


if __name__ == '__main__':
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python

#
# test_bgp_ecmp_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NETDEF DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NETDEF BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#

"""
test_bgp_ecmp_topo1.py: Test BGP topology with ECMP (Equal Cost MultiPath).
"""

import json
import functools
import os
import sys
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, '../'))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.
from mininet.topo import Topo

total_ebgp_peers = 20

#####################################################
#
#   Network Topology Definition
#
#####################################################


class BGPECMPTopo1(Topo):
    "BGP ECMP Topology 1"

    def build(self, **_opts):
        tgen = get_topogen(self)

        # Create the BGP router
        router = tgen.add_router('r1')

        # Setup Switches - 1 switch per 5 peering routers
        for swNum in range(1, (total_ebgp_peers + 4) / 5 + 1):
            switch = tgen.add_switch('s{}'.format(swNum))
            switch.add_link(router)

        # Add 'total_ebgp_peers' number of eBGP ExaBGP neighbors
        for peerNum in range(1, total_ebgp_peers+1):
            swNum = ((peerNum - 1) / 5 + 1)

            peer_ip = '10.0.{}.{}'.format(swNum, peerNum + 100)
            peer_route = 'via 10.0.{}.1'.format(swNum)
            peer = tgen.add_exabgp_peer('peer{}'.format(peerNum),
                                        ip=peer_ip, defaultRoute=peer_route)

            switch = tgen.gears['s{}'.format(swNum)]
            switch.add_link(peer)


#####################################################
#
#   Tests starting
#
#####################################################

def setup_module(module):
    tgen = Topogen(BGPECMPTopo1, module.__name__)
    tgen.start_topology()

    # Starting Routers
    router_list = tgen.routers()
    for rname, router in router_list.iteritems():
        router.load_config(
            TopoRouter.RD_ZEBRA,
            os.path.join(CWD, '{}/zebra.conf'.format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP,
            os.path.join(CWD, '{}/bgpd.conf'.format(rname))
        )
        router.start()

    # Starting Hosts and init ExaBGP on each of them
    topotest.sleep(10, 'starting BGP on all {} peers'.format(total_ebgp_peers))
    peer_list = tgen.exabgp_peers()
    for pname, peer in peer_list.iteritems():
        peer_dir = os.path.join(CWD, pname)
        env_file = os.path.join(CWD, 'exabgp.env')
        peer.start(peer_dir, env_file)
        logger.info(pname)


def teardown_module(module):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_convergence():
    "Test for BGP topology convergence"
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Expected result
    router = tgen.gears['r1']
    if router.has_version('<', '3.0'):
        reffile = os.path.join(CWD, 'r1/summary20.txt')
    else:
        reffile = os.path.join(CWD, 'r1/summary.txt')

    expected = json.loads(open(reffile).read())

    def _output_summary_cmp(router, cmd, data):
        """
        Runs `cmd` that returns JSON data (normally the command ends
        with 'json') and compare with `data` contents.
        """
        output = router.vtysh_cmd(cmd, isjson=True)
        if 'ipv4Unicast' in output:
            output['ipv4Unicast']['vrfName'] = \
                    output['ipv4Unicast']['vrfName'].replace(
                        'default', 'Default')
        elif 'vrfName' in output:
            output['vrfName'] = output['vrfName'].replace('default', 'Default')
        return topotest.json_cmp(output, data)

    test_func = functools.partial(
        _output_summary_cmp, router, 'show ip bgp summary json', expected)
    _, res = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assertmsg = 'BGP router network did not converge'
    assert res is None, assertmsg


def test_bgp_ecmp():
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    expect = {
        'routerId': '10.0.255.1',
        'routes': {
        },
    }

    for net in range(1, 5):
        for subnet in range(0, 10):
            netkey = '10.20{}.{}.0/24'.format(net, subnet)
            expect['routes'][netkey] = []
            for _ in range(0, 10):
                peer = {'multipath': True, 'valid': True}
                expect['routes'][netkey].append(peer)

    test_func = functools.partial(topotest.router_json_cmp,
                                  tgen.gears['r1'], 'show ip bgp json', expect)
    _, res = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assertmsg = 'expected multipath routes in "show ip bgp" output'
    assert res is None, assertmsg


if __name__ == '__main__':
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
